from .repostate import GitRepoState
from git.exc import InvalidGitRepositoryError, GitCommandError
from pathlib import Path
import os
import click


def pathlen(path):
    """Return folder level of the path string"""
    return len(Path(path).parts)


def reposearch(*folders, prune=False, level=0):
    for folder in folders:
        for dir, subdirs, files in os.walk(folder):
            try:
                yield (pathlen(dir) - pathlen(folder)), GitRepoState(dir)
                if prune:
                    subdirs[:] = []
                else:
                    subdirs[:] = [s for s in subdirs if s != '.git']
                continue
            except InvalidGitRepositoryError:
                pass


def ahead_behind(ref):
    return '{}/{}'.format(
        click.style('+' + str(ref['ahead']), fg='red') if ref['ahead'] else '-',
        click.style('-' + str(ref['behind']), fg='red') if ref['behind'] else '-')


def branch_representation(branch, brief=False):
    if branch['upstream']:
        if brief:
            refview = '{:.10} [{}]'.format(branch['name'], ahead_behind(branch))
        else:
            refview = '  Name: {:10.10} Upstream: {!s:17.17} Status: {!s}'.format(
                branch['name'], branch['upstream'], ahead_behind(branch))
    else:
        refview = '{:.10} {}' if brief else '  Name: {:10.10} Upstream: '
        refview = refview.format(branch['name'], click.style('None', fg='yellow'))
    return refview


def short_state_representation(repo):
    """Print the state for a repository"""
    loc = repo.path
    state = repo.state
    mod = click.style('Modified', fg='red') if state['dirty'] else click.style('Clean', fg='green')
    untracked = 'Untracked: '

    if len(state['untracked']) > 3:
        untracked += click.style('[ {}, {}, {}, ... ]'.format(*state['untracked'][:3]), fg='yellow')
    else:
        untracked += click.style(str(state['untracked']), fg='yellow')
    refs = 'Branches: '
    refs += ' '.join([branch_representation(b, brief=True) for b in state['refcheck']])
    return ' '.join([loc, mod, untracked, refs])


def branch_uptodate(branch, true_on_missing_origin=True):
    """Return True is branch is up to date with origin, otherwise False,
    also returns True if no remote defined"""
    if branch['upstream']:
        if branch['ahead'] or branch['behind']:
            return False
        return True
    if true_on_missing_origin:
        return True
    return False
    
def long_state_representation(repo):
    """Print the state for a repository"""
    loc = 'Location: ' + repo.path
    state = repo.state
    mod = 'Modified: {}'.format(click.style('Yes', fg='red') if state['dirty'] else click.style('No', fg='green'))
    untracked = 'Untracked: {}'.format(click.style(str(state['untracked']), fg='yellow'))
    refs = 'Branches: \n'
    refs += '\n'.join([branch_representation(b) for b in state['refcheck']])
    return '\n'.join([loc, mod, untracked, refs])


def repo_view(repo, brief=False):
    """Print all repo current state to terminal"""
    surrounds = '-' * 40 + '\n{}\n' + '-' * 40
    try:
        view = short_state_representation(repo) if brief else surrounds.format(long_state_representation(repo))
    except GitCommandError as gce:
        if 'must be run in a work tree' in gce.stderr:
            if brief:
                loc = repo.path
                msg = click.style('Bare repo', fg='yellow')
                view = ' '.join([loc, msg])
            else:
                loc = 'Location: ' + repo.path
                msg = click.style('Repository is bare and does not have state to track', fg='yellow')
                view = surrounds.format('\n'.join([loc, msg]))
    return view



from git import Repo
from git.exc import InvalidGitRepositoryError
from colorama import Fore, Style, init
import os
import warnings

init()

class GitRepoState(Repo):
    """An object that represents the state of a repository"""
    def __init__(self, directory, **kwargs):
        self.path = os.path.abspath(directory)
        super().__init__(self.path, **kwargs)

    def check_upstream(self, ref):
        """replicate my for-each-ref pattern"""
        upstream = ref.tracking_branch()
        if not upstream:
            return dict(name=ref.name, upstream=None)
        ahead = sum(1 for c in
                    self.iter_commits(rev=upstream.name + '..' + ref.name))
        behind = sum(1 for c in
                     self.iter_commits(rev=ref.name + '..' + upstream.name))
        return dict(name=ref.name, upstream=upstream, ahead=ahead,
                    behind=behind)

    def _checkall_upstreams(self):
        """Check a git repo state and report it"""
        return [self.check_upstream(ref) for ref in self.heads]

    @property
    def state(self):
        """Return the current state for this repository"""
        return {
            'dirty': self.is_dirty(),
            'untracked': self.untracked_files,
            'refcheck': self._checkall_upstreams()
            }
        



from repovisor import repovisor as rv
import click


@click.group()
@click.version_option()
def main():
    pass


@main.command(name='search')
@click.argument('folders', nargs=-1, type=click.Path(exists=True, file_okay=False, resolve_path=True))
@click.option('--brief', '-b', is_flag=True, help='Use the short form of reporting status')
@click.option('--prune', '-p', is_flag=True, help='Prune directories as soon as we reach a repository')
@click.option('--tree', '-t', is_flag=True, help='Show repositories in a tree structure, only used if brief is also selected')
@click.option('--hide', is_flag=True, help='Hide repositories that are clean')
@click.option('--remote', is_flag=True, help='Warn if repo has no remote')
def __search_action(folders, brief, prune, tree, hide, remote):
    '''search through a set of folders and print status for them'''
    for repo_level, repo in rv.reposearch(*folders, prune=prune, level=0):
        view = rv.repo_view(repo, brief=brief)
        if tree and brief:
            view = '-'*repo_level + view
        if hide and not repo.bare:
            if not repo.state['dirty']:
                if not repo.state['untracked']:
                    if all(rv.branch_uptodate(b, true_on_missing_origin=not remote) for b in repo.state['refcheck']):
                        continue
        print(view)
